/*
 ============================================================================
 Name        : MyersMillerAlign.c
 Author      : Shahidul Islam
 Version     : 0.2 (09.27.2013 - 09.30.2013)
 	 	 	   0.1 (01.27.2013 - 02.01.2013)
 Copyright   : Your copyright notice
 Description : Optimal alignments in linear space, proposed by E.W.Myers and W.Miller

Modified Features (Version: 0.2):
	1. Read sequence from 2 different files and compare them.
	2. Handle FASTA format.
		- Accepting A,T,C,G,U,N character.
		- Change U -> T and N -> Any Random Sequence

Features (Version: 0.1):
    1. Implemented focusing DNA/RNA sequences, and will show explicit warning for characters other than A/T/C/G/U.
       To get rid of the warnings for any sequence, global enum 'sequence_type' need to be set as 'ANY'
    2. The required memory = (2 * Length of 1st sequence) + (6 * Length of 2nd sequence) + A Few Scalars
    3. Gap penalty has been considered.
    4. Penalty and score matrix has been used according to the examples of the article.
    5. Many possible invalid input has been explicitly addressed.
    6. A self testing module has been added to detect if the proposed alignment actually generates
       optimal score or not.

 ============================================================================
 */


#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<ctype.h>
#include<time.h>
#include "./inc/MyersMillerAlign.h"
#include "./inc/UtilityFunctions.h"

//Values for g,h has been defined, according to the examples of the article
#define G 11//3//11//8.0//2
#define H 1//0.3//1//0.2//0.5
#define REPLACE_COST_DUMMY 3
#define REDUCED_PERCENT 1.1 //DUMMY

//Memory reallocation manage with chunk
#define MEMORY_CHUNK 4000

//Maximum number of characters to display in a row of an alignment
#define ALIGN_WIDTH 60

#define FILE_NAME_SIZE 128
#define INPUT_FILE_LINE_MAX_LENGHT 120

//State to generate warning for non-nucleotide symbols, when expected sequence is DNA/RNA
enum
{
    DNA_RNA,
    ANY
}sequence_type = DNA_RNA;

int takenInputDummy = 1;
double replaceCost = 3;//0.6;//0.5;//1.1;
double reducedCost = 2.7;//0.6;//0.5;//0.9;
int printAlignment = 0;

int MMAlign(char seq1[], int m, char seq2[], int n, char *state, double *cost);

//Functions to generate, print and test the alignment
void diff(char *a, char *b, int m, int n, double *cc, double *dd, double *rr, double*ss, double **w, char *state, double tb, double te);
double cost_only_in_linear_space(char *a, char *b, int m, int n, double *ccc, double *ddd, double **w, double g, int direction);
void align_place(char *state, int value, char a, char b);
void print_alignment(char *a, char *b, int m, int n, char *state);
double gap(int k);
double cost_calculation_from_alignment(char *state);

//Functions to intialize, get input and test input consistency
void initialize_W(double **w, int size);
double replaceCostFunc(char c1, char c2);
int is_valid_nucleotide(char *c);
void random_nucleotide(char *c);

//Read Fasta File
int readFastaFiles(char input_file[], sequenceMap *refSeqMap);
void processHeader(char *line, char pdb1[], char chain1[], char pdb2[], char chain2[], sequenceMap refSeqMap[], int flag[]);
int addToSequence(char *line, chainMap *chainSeq);

void alignFasta(sequenceMap refSeqMap[], sequenceMap refMap[], sequenceMap pdbTempMap[], basePairInfo basePairMapping[][MAX_BASE_PAIR], int basePairCountsMapping[])
{
	printAlignment = 0;
	align_length = 0;
	//m and n is size of the sequences
    int i; //m = 0, n = 0, max_number_of_char = 128;

	strcpy(refSeqMap[0].aChainMap[refSeqMap->chainCount - 1].mappedChainId, refSeqMap[1].aChainMap[refSeqMap->chainCount - 1].chainId);
	strcpy(refSeqMap[1].aChainMap[refSeqMap->chainCount - 1].mappedChainId, refSeqMap[0].aChainMap[refSeqMap->chainCount - 1].chainId);

	for(i = 0; i < min(refSeqMap[0].chainCount, refSeqMap[1].chainCount); i++)
		chainAlignAnchorBased(&(refSeqMap[0].aChainMap[i]), &(refSeqMap[1].aChainMap[i]), refMap, pdbTempMap, basePairMapping, basePairCountsMapping);

	if(PRINT_SMALL_FINAL_INFO)
		printf("Sequence Lengths = %d and %d\n", refSeqMap[0].aChainMap[0].resCount, refSeqMap[1].aChainMap[0].resCount);
	printAlignment = 0;
}

void alignPDBWithRef(sequenceMap refSeqMap /*ref to ref map*/, atomInfo atomMList[], int atomMCount, sequenceMap *refMap/*ref to pdb map*/, sequenceMap *pdbMap)
{
	int i, j, resInd, prevResId[MAX_CHAIN_COUNT];
	char  prevICode[MAX_CHAIN_COUNT];

	for(i = 0; i < MAX_CHAIN_COUNT; i++)
	{
		prevResId[i] = -1;
		prevICode[i] = '\0';
	}

	//Copy ref sequence from fasta map data
	*pdbMap = *refMap = refSeqMap;
	strcpy(pdbMap->seqId, "ATOM");
	for(i = 0; i < refMap->chainCount; i++)
		pdbMap->aChainMap[i].resCount = 0;

	//Copy PDB sequence from atom info
	for(i = 0; i < atomMCount; i++)
	{
		for(j = 0; j < pdbMap->chainCount; j++)
		{
			if(strcmp(pdbMap->aChainMap[j].chainId, atomMList[i].chainId) == 0)
			{
				if(atomMList[i].resSeqNo != prevResId[j] || atomMList[i].iCode != prevICode[j])
				{
					resInd = pdbMap->aChainMap[j].resCount;
					pdbMap->aChainMap[j].resMap[resInd].residue = atomMList[i].residueName;
					pdbMap->aChainMap[j].resMap[resInd].resPos = atomMList[i].resSeqNo;
					pdbMap->aChainMap[j].resMap[resInd].iCode = atomMList[i].iCode;
					pdbMap->aChainMap[j].resCount++;
					prevResId[j] = atomMList[i].resSeqNo;
					prevICode[j] = atomMList[i].iCode;
				}
				break;
			}
		}
	}

	for(i = 0; i < pdbMap->chainCount; i++)
		chainAlign(&refMap->aChainMap[i], &pdbMap->aChainMap[i]);
}

void insertAnchorItem(anchorMap anchorList[MAX_HELIX_COUNT], int *anchorCount, int refStartBpInd, int targetStartBpInd, int length)
{
	anchorList[*anchorCount].pdb1IndRange[0] = refStartBpInd;
	anchorList[*anchorCount].pdb1IndRange[1] = refStartBpInd + length - 1;
	anchorList[*anchorCount].pdb2IndRange[0] = targetStartBpInd;
	anchorList[*anchorCount].pdb2IndRange[1] = targetStartBpInd + length - 1;
	//printf("Anchoring: %d %d -> %d %d\n", anchorList[*anchorCount].pdb1IndRange[0], anchorList[*anchorCount].pdb1IndRange[1], anchorList[*anchorCount].pdb2IndRange[0], anchorList[*anchorCount].pdb2IndRange[1]);
	//getchar();
	(*anchorCount)++;
}

int compAnchor(const void*A, const void*B)
{
	anchorMap *x = (anchorMap*)A;
	anchorMap *y = (anchorMap*)B;
	if(x->pdb1IndRange[0] == y->pdb1IndRange[0])
		return x->pdb1IndRange[1] - y->pdb1IndRange[1];
	else
		return x->pdb1IndRange[0] - y->pdb1IndRange[0];

}

void getAnchor(helix helixListRef[MAX_HELIX_COUNT], int helixCount, basePairInfo basePairMapping[][MAX_BASE_PAIR], int basePairCountsMapping[], int minHelixLen, anchorMap anchorList[MAX_HELIX_COUNT], int *anchorCount)
{
	*anchorCount = 0;
	int printSteps = 0;
	int windowSize = 20;
	int i, j, k, l, start = max(helixListRef[0].basePairIndex[0] - windowSize / 2, 0);
	int matchCount, startInd, matchStart, maxMatchCount, maxStartInd;
	int seqPosToBpPosMap[MAX_SEQ_LEN];
	char pdb1iCodes[2], pdb2iCodes[2];

	//seqPos to BpPos Mapping
	for(j = 0; j <= basePairMapping[1][0].residueID[0]; j++)
		seqPosToBpPosMap[j] = 0;
	for(i = 0; i < basePairCountsMapping[1] - 1; i++)
	{
		for(j = basePairMapping[1][i].residueID[0] + 1; j <= basePairMapping[1][i + 1].residueID[0]; j++)
			seqPosToBpPosMap[j] = i + 1;
	}

	basePairInfo prevBP;
	for(i = 0; i < helixCount; i++)
	{
		//Revise this range of search
		int helix1stInd = basePairMapping[0][helixListRef[i].basePairIndex[0]].residueID[0];
		int helixMappedPos = seqPosToBpPosMap[helix1stInd];
		if(start < helixMappedPos - windowSize / 2)
			start = max(helixMappedPos - windowSize / 2, 0);
		if(printSteps)
		{
			printf("Window: %d %d\n", start, helixMappedPos + windowSize/2);
			getchar();
		}
		for(j = start; j < helixMappedPos + windowSize/2 && j < basePairCountsMapping[1]; j++)
        {
			//if(i > 0 && basePairMapping[1][j].residueID[0] > 18)//helixListRef[i].basePairIndex[0] - 2)
				//printSteps = 1;
			//printf("Step1: %d %d\n", basePairMapping[1][j].residueID[0], helixListRef[i].basePairIndex[0] - 2);
			if(!isCanonicalBP(basePairMapping[1][j]))
				continue;
            matchCount = maxMatchCount = 0;
            maxStartInd = j;
            for(k = helixListRef[i].basePairIndex[0], l = j; k <= helixListRef[i].basePairIndex[1] && l < basePairCountsMapping[1]; )
            {
            	if(printSteps)
            	{
            		printf("j = %d, k = %d, l = %d\n", j, k, l);
					printPairInfoOnScreen(basePairMapping[0][k]);
					printPairInfoOnScreen(basePairMapping[1][l]);
            	}

            	//Skip non-canonical basepairs
            	if(!isCanonicalBP(basePairMapping[0][k]) || !isCanonicalBP(basePairMapping[1][l]))
            	{
            		if(!isCanonicalBP(basePairMapping[0][k]))
            			k++;
            		if(!isCanonicalBP(basePairMapping[1][l]))
            			l++;
            	}
            	else
            	{
					//Compare base pairs
					if(basePairMapping[1][l].residueID[0] != prevBP.residueID[0] + 1 || basePairMapping[1][l].residueID[1] != prevBP.residueID[1] - 1 ||
						compBasePairAndEdge(&basePairMapping[0][k], &basePairMapping[1][l]) != 0)
					{
						matchCount = 0;
						l = j - 1;
					}
					else
						matchCount++;

					if(matchCount == 1)
						startInd = k;

					if(matchCount > maxMatchCount)
					{
						maxMatchCount = matchCount;
						maxStartInd = startInd;

						if(printSteps)
							if(maxMatchCount >= minHelixLen)
								printf("Helix Match Found j = %d, %d %d\n", j, maxMatchCount, maxStartInd);
					}

					if(printSteps)
					{
						printf("Match = %d\n", matchCount);
						getchar();
					}
					prevBP = basePairMapping[1][l];
					k++;
					l++;
            	}
            }
            if(maxMatchCount >= minHelixLen)
            {
                insertAnchorItem(anchorList, anchorCount, basePairMapping[0][maxStartInd].residueID[0], basePairMapping[1][j].residueID[0], maxMatchCount);
                insertAnchorItem(anchorList, anchorCount, basePairMapping[0][maxStartInd].residueID[1] - maxMatchCount + 1, basePairMapping[1][j].residueID[1] - maxMatchCount + 1, maxMatchCount);
                start = j + maxMatchCount + 1;
                break;
            }
        }
	}
	qsort(anchorList, *anchorCount, sizeof(anchorMap), compAnchor);
	/*for(i = 0; i < *anchorCount; i++)
		printf("Anchoring: %4d %4d -> %4d %4d\n", anchorList[i].pdb1IndRange[0], anchorList[i].pdb1IndRange[1], anchorList[i].pdb2IndRange[0], anchorList[i].pdb2IndRange[1]);
	printf("Total Anchor = %d\n", *anchorCount);
	getchar();*/
}

//Send pdbMap[refPDBIndex].aChainMap[j].resMap, pdbMap[refPDBIndex].aChainMap[j].resCount
//Handle the -1
int getPDBresIdToArrayInd(int resPos, char iCode, posMap resMapList[], int resCount)
{
	posMap *aPos = NULL, tempPos;
	tempPos.resPos = resPos;
	tempPos.iCode = iCode;
	aPos = (posMap *)bsearch(&tempPos, resMapList, resCount, sizeof(posMap), mappedSeqComp);

	if(aPos)
		return aPos - resMapList;
	else
	{
		printf("Error!! Base pair generating method has generated pairs with invalid '%c'%d\n", resPos, iCode);
		return -1;
	}
}

void chainAlignAnchorBased(chainMap *seqAlign1, chainMap *seqAlign2, sequenceMap refMap[], sequenceMap pdbTempMap[], basePairInfo basePairMapping[][MAX_BASE_PAIR], int basePairCountsMapping[])
{
	int printDetails = 0;
	//int minHelixLen = 4;

	int ind, startAnchor = 0;

	anchorCount = 0;

	int i, j, k, count1, count2;
	sequence1 = sequence2 = state = NULL;
	cc = dd = rr = ss = NULL;
	w = NULL;

	sequence1 = (char *) malloc((seqAlign1->resCount + 1) * sizeof(char));
	sequence2 = (char *) malloc((seqAlign2->resCount + 1) * sizeof(char));

	state = (char *)malloc((seqAlign1->resCount + seqAlign2->resCount) * sizeof(char));

	int start1 = 0, start2 = 0;
	int end1, end2, end1Lim = seqAlign1->resCount, end2Lim = seqAlign2->resCount;

	seqAlign1->resCount = seqAlign2->resCount = 0;

	double alignCost, totalAlignCost = 0.0;

	for(k = startAnchor; k <= anchorCount; k++)
	{
		alignCost = 0.0;
		align_length = 0;
		if(k == anchorCount)
		{
			end1 = end1Lim;
			end2 = end2Lim;
		}
		else
		{
			//Considering the PDB is holding only one chain
			//Update With iCode
			ind = getPDBresIdToArrayInd(anchorList[k].pdb1IndRange[0], ' ', pdbTempMap[0].aChainMap[0].resMap, pdbTempMap[0].aChainMap[0].resCount);
			if(ind >= 0)
				end1 = pdbTempMap[0].aChainMap[0].resMap[ind].seq2MappedPos;
			else
			{
				printf("Error1!!\n");
				getchar();
			}
			ind = getPDBresIdToArrayInd(anchorList[k].pdb2IndRange[0], ' ', pdbTempMap[1].aChainMap[0].resMap, pdbTempMap[1].aChainMap[0].resCount);
			if(ind >= 0)
				end2 = pdbTempMap[1].aChainMap[0].resMap[ind].seq2MappedPos;
			else
			{
				printf("Error2!!\n");
				getchar();
			}
		}

		if(printDetails)
		{
			printf("Start1 = %d End1 = %d Start2 = %d End2 = %d\n", start1, end1, start2, end2);
			getchar();
		}

		if(start1 == end1 || start2 == end2)
			continue;

		for(i = start1, count1 = 0; i < end1; i++, count1++)
			sequence1[count1] = seqAlign1->resMap[i].residue;
		sequence1[count1] = 0;

		for(i = start2, count2 = 0; i < end2; i++, count2++)
			sequence2[count2] = seqAlign2->resMap[i].residue;
		sequence2[count2] = 0;

		MMAlign(sequence1, count1, sequence2, count2, state, &alignCost);
		totalAlignCost += alignCost;

		for(i = 0; i < align_length; i++)
		{
			if(state[i] == 'x')
			{
				seqAlign1->resMap[seqAlign1->resCount].isMatch = 0;
				seqAlign1->resMap[seqAlign1->resCount].seq2MappedPos = NO_ALIGN_INDEX;
				seqAlign1->resCount++;
			}
			else if(state[i] == '-')
			{
				seqAlign2->resMap[seqAlign2->resCount].isMatch = 0;
				seqAlign2->resMap[seqAlign2->resCount].seq2MappedPos = NO_ALIGN_INDEX;
				seqAlign2->resCount++;
			}
			else
			{
				if(state[i] == '|')
				{
					seqAlign1->resMap[seqAlign1->resCount].isMatch = 0;
					seqAlign2->resMap[seqAlign2->resCount].isMatch = 0;
				}
				else
				{
					seqAlign1->resMap[seqAlign1->resCount].isMatch = 1;
					seqAlign2->resMap[seqAlign2->resCount].isMatch = 1;
				}

				//The mapped position is sequential array position
				//It is required that the position used as index to get the actual PDB position
				//seqAlign1->resMap[seqAlign1->resCount].resPos = seqAlign1->resCount;
				seqAlign1->resMap[seqAlign1->resCount].seq2MappedPos = seqAlign2->resCount;
				//seqAlign2->resMap[seqAlign2->resCount].resPos = seqAlign2->resCount;
				seqAlign2->resMap[seqAlign2->resCount].seq2MappedPos = seqAlign1->resCount;
				seqAlign1->resCount++;
				seqAlign2->resCount++;
			}
		}
		if(k != anchorCount)
		{
			/*if(seqAlign1->resCount != pdbTempMap[0].aChainMap[0].resMap[anchorList[k].pdb1IndRange[0]].seq2MappedPos ||
				seqAlign2->resCount != pdbTempMap[1].aChainMap[0].resMap[anchorList[k].pdb2IndRange[0]].seq2MappedPos)
			{
				printf("%d %d %d %d\n", seqAlign1->resCount, anchorList[k].pdb1IndRange[0], seqAlign2->resCount, anchorList[k].pdb2IndRange[0]);
				getchar();
			}*/
			for(i = anchorList[k].pdb1IndRange[0], j = anchorList[k].pdb2IndRange[0]; i <= anchorList[k].pdb1IndRange[1]; i++, j++)
			{
				seqAlign1->resMap[seqAlign1->resCount].isMatch = 1;
				seqAlign2->resMap[seqAlign2->resCount].isMatch = 1;
				seqAlign1->resMap[seqAlign1->resCount].seq2MappedPos = seqAlign2->resCount;
				seqAlign2->resMap[seqAlign2->resCount].seq2MappedPos = seqAlign1->resCount;
				seqAlign1->resCount++;
				seqAlign2->resCount++;
			}
			ind = getPDBresIdToArrayInd(anchorList[k].pdb1IndRange[1], ' ', pdbTempMap[0].aChainMap[0].resMap, pdbTempMap[0].aChainMap[0].resCount);
			if(ind >= 0)
				start1 = pdbTempMap[0].aChainMap[0].resMap[ind].seq2MappedPos + 1;
			else
			{
				printf("Error3!!\n");
				getchar();
			}
			ind = getPDBresIdToArrayInd(anchorList[k].pdb2IndRange[1], ' ', pdbTempMap[1].aChainMap[0].resMap, pdbTempMap[1].aChainMap[0].resCount);
			if(ind >= 0)
				start2 = pdbTempMap[1].aChainMap[0].resMap[ind].seq2MappedPos + 1;
			else
			{
				printf("Error4!!\n");
				getchar();
			}
		}
	}
	if(PRINT_ALIGN_BP_INFO)
		printf("Total Align Cost = %0.2lf\n", totalAlignCost);
	//getchar();

	if(sequence1)
		free(sequence1);
	if(sequence2)
		free(sequence2);
	if(state)
		free(state);
	return;
}

void chainAlign(chainMap *seqAlign1, chainMap *seqAlign2)
{
	int i;
	sequence1 = sequence2 = state = NULL;
	cc = dd = rr = ss = NULL;
	w = NULL;
	align_length = 0;
	sequence1 = (char *) malloc((seqAlign1->resCount + 1) * sizeof(char));
	sequence2 = (char *) malloc((seqAlign2->resCount + 1) * sizeof(char));

	for(i = 0; i < seqAlign1->resCount; i++)
		sequence1[i] = seqAlign1->resMap[i].residue;
	sequence1[seqAlign1->resCount] = 0;

	for(i = 0; i < seqAlign2->resCount; i++)
		sequence2[i] = seqAlign2->resMap[i].residue;
	sequence2[seqAlign2->resCount] = 0;

	state = (char *)malloc((seqAlign1->resCount + seqAlign2->resCount) * sizeof(char));

	double alignCost;
	MMAlign(sequence1, seqAlign1->resCount, sequence2, seqAlign2->resCount, state, &alignCost);

	seqAlign1->resCount = seqAlign2->resCount = 0;
	for(i = 0; i < align_length; i++)
	{
		if(state[i] == 'x')
		{
			seqAlign1->resMap[seqAlign1->resCount].isMatch = 0;
			seqAlign1->resMap[seqAlign1->resCount].seq2MappedPos = NO_ALIGN_INDEX;
			seqAlign1->resCount++;
		}
		else if(state[i] == '-')
		{
			seqAlign2->resMap[seqAlign2->resCount].isMatch = 0;
			seqAlign2->resMap[seqAlign2->resCount].seq2MappedPos = NO_ALIGN_INDEX;
			seqAlign2->resCount++;
		}
		else
		{
			if(state[i] == '|')
			{
				seqAlign1->resMap[seqAlign1->resCount].isMatch = 0;
				seqAlign2->resMap[seqAlign2->resCount].isMatch = 0;
			}
			else
			{
				seqAlign1->resMap[seqAlign1->resCount].isMatch = 1;
				seqAlign2->resMap[seqAlign2->resCount].isMatch = 1;
			}

			//The mapped position is sequential array position
			//It is required that the position used as index to get the actual PDB position
			//seqAlign1->resMap[seqAlign1->resCount].resPos = seqAlign1->resCount;
			seqAlign1->resMap[seqAlign1->resCount].seq2MappedPos = seqAlign2->resCount;
			//seqAlign2->resMap[seqAlign2->resCount].resPos = seqAlign2->resCount;
			seqAlign2->resMap[seqAlign2->resCount].seq2MappedPos = seqAlign1->resCount;
			seqAlign1->resCount++;
			seqAlign2->resCount++;
		}
	}

	if(sequence1)
		free(sequence1);
	if(sequence2)
		free(sequence2);
	if(state)
		free(state);
	return;
}

//Process the header of FASTA format to find the index of PDB in the list that match with the current header
//If no match found, return -1
int processHeaderList(char *line, char pdbs[][PDB_ID_SZ], char chains[][MAX_CHAIN_NAME_LENGTH], int count, char seq[][MAX_SEQ_LEN])
{
	int i, len = strlen(line);
	
	if(len < 7)
		return -1;
	
	for(i = 0; i < 7; i++)
		line[i] = toupper(line[i]);

	for(i = 0; i < count; i++)
	{
		if(strncmp(&line[1], pdbs[i], 4) == 0 && strncmp(&line[6], chains[i], strlen(chains[i])) == 0)
			return i;
	}
	return -1;
}

//For a given list of PDBs, read the fasta sequence
int readFastaSeqList(char input_file[], char pdbs[][PDB_ID_SZ], char chains[][MAX_CHAIN_NAME_LENGTH], int count, char seq[][MAX_SEQ_LEN])
{
	char line[MAX_SEQ_LEN + 1];
	int flag = 0, gotSequence = 0, index = -1;

	int k;
	for(k = 0; k < count; k++)
		strcpy(seq[k], "");
	
	FILE *fpt = fopen(input_file, "r");
	if(!fpt)
	{
		printf("File \"%s\" was not found.\n", input_file);
		return 0;
	}
	else
	{
		while(fscanf(fpt," %[^\n]",line) == 1)
		{
			if(line[0] == '>')
				index = processHeaderList(line, pdbs, chains, count, seq);
			else
			{
				int i;				
				if(index >= 0)
				{
					gotSequence++;
					strcpy(seq[index], line);
				}
				index = -1;
			}
		}
	}
	if(fpt)
		fclose(fpt);
	if(gotSequence < 1)
		return 0;
	return 1;
}

int readFastaSequence(char input_file[], char pdb1[], char chain1[], char pdb2[], char chain2[], sequenceMap refSeqMap[])
{
	//I/O file handling
	char line[MAX_SEQ_LEN + 1];
	int flag[2] = {0, 0}, gotSequence = 0;
	refSeqMap[0].chainCount = refSeqMap[1].chainCount = 0;

	FILE *fpt = fopen(input_file, "r");
	if(!fpt)
	{
		printf("File \"%s\" was not found.\n", input_file);
		return 0;
	}
	else
	{
		while(fscanf(fpt," %[^\n]",line) == 1)
		{
			if(line[0] == '>')
				processHeader(line, pdb1, chain1, pdb2, chain2, refSeqMap, flag);
			else
			{
				int i;
				for(i = 0; i <= 1; i++)
				{
					if(flag[i])
					{
						gotSequence++;
						if(refSeqMap[i].chainCount > 0)
						{
							addToSequence(line, &refSeqMap[i].aChainMap[refSeqMap[i].chainCount - 1]);
						}
						flag[i] = 0;
					}
				}
			}
		}
	}
	if(fpt)
		fclose(fpt);
	if(gotSequence < 2)
		return 0;
	return 1;
}

//Process the first header of FASTA format
void processHeader(char *line, char pdb1[], char chain1[], char pdb2[], char chain2[], sequenceMap refSeqMap[], int flag[])
{
	int i, len = strlen(line);
	if(len < 8)
		return;
	
	int chainLen[2];
	chainLen[0] = strlen(chain1);
	chainLen[1] = strlen(chain2);

	for(i = 0; i < 8; i++)
		line[i] = toupper(line[i]);

	for(i = 0; i < 2; i++)
	{
		if((i == 0 && strncmp(pdb1, &line[1], 4) == 0 && strncmp(&line[6], chain1, chainLen[0]) == 0) ||
			(i == 1 && strncmp(pdb2, &line[1], 4) == 0 && strncmp(&line[6], chain2, chainLen[1]) == 0))
		{
			strncpy(refSeqMap[i].seqId, &line[1], 4);
			strncpy(refSeqMap[i].aChainMap[refSeqMap[i].chainCount].chainId, &line[6], chainLen[i]);
			refSeqMap[i].aChainMap[refSeqMap[i].chainCount].resCount = 0;
			refSeqMap[i].chainCount++;
			flag[i] = 1;
			if(PRINT_SMALL_FINAL_INFO)
				printf("%s - %s\n", refSeqMap[i].seqId, refSeqMap[i].aChainMap[refSeqMap[i].chainCount - 1].chainId);
		}
	}
}

/*
- Take 2 sequences as input, test validity, and give warning if necessary.
- Fix necessary inconsistencies to continue calculations, within allocated memory.
- Return 1 if there are some inconsistency, 0 otherwise.
- If the calculation should be blocked for any inconsistency, the function should return '-1' for those cases.
*/

int addToSequence(char *line, chainMap *chainSeq)
{
	int ret = 0, i, len = strlen(line);
	//char ch;
	//Take the 1st sequence as input
	for(i = 0; i < len; i++)
	{
		//Test and give warning if there are less characters than defined as length of 1st sequence
		if(is_whitespace(line[i]))
		{
			printf("No valid nucleotide in seq-1 at position-%d\n", chainSeq->resCount + i);
			ret = 1;
		}
		//Test if expected sequences are DNA/RNA, and generate warning for non-ACGTU characters
		else if(sequence_type == DNA_RNA && !is_valid_nucleotide(&line[i]))
		{
			printf("Invalid nucleotide \'%c\' in seq-1 at position-%d\n", line[i], chainSeq->resCount + i);
			ret = 1;
		}
	}
	//Give an overall warning to draw attention to various inconsistencies
	if(ret)
		printf("--Warning!!! There are data inconsistency, as described above--\n--Continuing with invalid data--\n\n");

	for(i = 0; i < len; i++)
	{
		chainSeq->resMap[chainSeq->resCount].resPos = chainSeq->resCount + 1;
		chainSeq->resMap[chainSeq->resCount].residue = line[i];
		chainSeq->resMap[chainSeq->resCount].iCode = NO_ICODE;
		chainSeq->resCount++;
		//printf("%c %d", chainSeq->resMap[chainSeq->resCount - 1].residue, chainSeq->resMap[chainSeq->resCount - 1].resPos);
		//getchar();
	}

	return ret;
}

int MMAlign(char seq1[], int m, char seq2[], int n, char *state, double *cost)
{
	int i, max_number_of_char = 128;

    //Initialization for generating random numbers
	srand(time(NULL));

	//Memory allocation for the score matix of size*size
	w = (double **)malloc(max_number_of_char * sizeof(double));
	if(!w)
	{
		printf("Sufficient memory is not available to process the data\n");
		return 0;
	}
	for(i = 0; i < max_number_of_char; i++)
	{
		w[i] = (double *)malloc(max_number_of_char * sizeof(double));
		if(!w[i])
		{
			printf("Sufficient memory is not available to process the data\n");
			return 0;
		}
	}

	initialize_W(w, max_number_of_char);
	
	//Memory allocations for vectors
	cc = (double *)malloc((n + 1) * sizeof(double));
	dd = (double *)malloc((n + 1) * sizeof(double));
	rr = (double *)malloc((n + 1) * sizeof(double));
	ss = (double *)malloc((n + 1) * sizeof(double));


	if(!cc || !dd || !rr || !ss || !state)
	{
		printf("Sufficient memory is not available to process the data\n");
		return 0;
	}

	//Call the recursion of divide and conquer
	if(!doScoreOnlyAlign)
		diff(seq1, seq2, m, n, cc, dd, rr, ss, w, state, G, G);

	//To test the accuracy of the algorithm: Part - 1
	//Calculate the alignment using cost only module
	*cost = cost_only_in_linear_space(seq1, seq2, m, n, cc, dd, w, G, 1);

	//Print Alignment
	if(printAlignment)
	{
		if(PRINT_ALIGN_BP_INFO)
			printf("Optimal Cost = %0.2lf\n\n", *cost);
		print_alignment(seq1, seq2, m, n, state);
		getchar();
	}

	//To test the accuracy of the algorithm: Part - 2
	//Calculate the cost from proposed alignment and match with optimal
	/*double cost2 = cost_calculation_from_alignment(state);
	if(myFabs(cost2 - cost)>0.000001)
	{
		printf("Divide & Conquer Algorithm Alignment Cost = %0.2lf\n", cost2);
		printf("----- ERROR!!! ------\n");
		printf("-- This program does not propose the optimal alignment for these 2 sequences --\n");
		printf("-- It would be a great help if you mail these 2 sequences to: si_sumon@yahoo.com --\n");
	}*/

	//Free all allocated memory

	if(cc)      free(cc);
	if(dd)      free(dd);
	if(rr)      free(rr);
	if(ss)      free(ss);

	if(w)
	{
		for(i = 0; i < max_number_of_char; i++)
			if(w[i]) free(w[i]);
		free(w);
	}

	return 0;
}

//The core module to implement the divide and conquer algorithm for delivering a conversion in linear space
void diff(char *a, char *b, int m, int n, double *cc, double *dd, double *rr, double*ss, double **w, char *state, double tb, double te)
{
    if(n == 0)
    {
        //Delete 'a', in base condition of n==0
        if(m > 0)
            align_place(state, -m, ' ', ' ');
    }
    else if(m == 0)
        //Insert 'b', in base condition of m==0
        align_place(state, n, ' ', ' ');
    //Another base condition handling, where m==1
    else if(m == 1)
    {
        //Initialize 'min_val' with cost of inserting 'b' and deleting 'a[1]'
        double min_val = min(tb, te) + H + gap(n), val;
        int  min_ind, flag = 0;
        int a1 = a[0], bj, j;

        //Look for better alignment with one replace operation
        for(j = 1; j <= n; j++)
        {
            bj = b[j-1];
            val = gap(j - 1) + w[a1][bj] + gap(n - j);
            if((!flag && val <= min_val) || (flag && val <= min_val))
            {
                flag = 1;
                min_val = val;
                min_ind = j - 1;
            }
        }

        //'flag == 0' means that the optimal operations contains a delete and an insert
        if(flag == 0)
        {
            //If tb is '0', then the first operation is delete
            if(tb == 0)
            {
                if(n)   align_place(state, -m, ' ', ' ');
                if(m)   align_place(state, n, ' ', ' ');
            }
            //If tb is not '0', then the first operation is insert
            else
            {
                if(n)   align_place(state, n, ' ', ' ');
                if(m)   align_place(state, -m, ' ', ' ');
            }
        }
        else
        {
            //Insert-Replace-Insert operations (skips the 0 item insert operation)
            if(min_ind)
                align_place(state, min_ind, ' ', ' ');
            align_place(state, 0, a[0], b[min_ind]);
            if(n - (min_ind + 1))
                align_place(state, n-(min_ind + 1), ' ', ' ');
        }
    }
    //Dividing the problem, and conquering
    else
    {
        //Middle row
        int i_opt = m/2;

        //Forward (Last Parameter = '1') application of cost calculating module
        cost_only_in_linear_space(a, b, i_opt, n, cc, dd, w, tb, 1);
        //Reverse (Last Parameter = '-1') application of cost calculating module
        cost_only_in_linear_space(&a[i_opt], b, m - i_opt, n, rr, ss, w, te, -1);
        int type, j = 0, j_opt = 0;
        double min_val;

        //Intialize the min_val and type
        if(cc[j] + rr[n-j] < dd[j] + ss[n-j] - G)
        {
            min_val = cc[j] + rr[n-j];
            type = 1;
        }
        else
        {
            min_val = dd[j] + ss[n-j] - G;
            type = 2;
        }

        //Look for the 'midpoint' of optimal alignment
        for(j = 1; j <= n; j++)
        {
            if(min(cc[j] + rr[n-j], dd[j] + ss[n-j] - G) <= min_val)
            {
                j_opt = j;
                if(cc[j] + rr[n-j] < dd[j] + ss[n-j] - G)
                {
                    min_val = cc[j] + rr[n-j];
                    type = 1;
                }
                else
                {
                    min_val = dd[j] + ss[n-j] - G;
                    type = 2;
                }
            }
        }

        //Split accordingly if cc[j_opt] + rr[n-j_opt] gives optimal
        if(type == 1)
        {
            diff(a, b, i_opt, j_opt, cc, dd, rr, ss, w, state, tb, G);
            diff(&a[i_opt], &b[j_opt], m - i_opt, n - j_opt, cc, dd, rr, ss, w, state, G, te);
        }
        //Split accordingly if dd[j_opt] + ss[n-j_opt] - G is optimal
        else
        {
            diff(a, b, i_opt - 1, j_opt, cc, dd, rr, ss, w, state, tb, 0);
            //delete a[i_opt-1] and a[i_opt]
            align_place(state, -2, ' ', ' ');
            diff(&a[i_opt + 1], &b[j_opt], m - (i_opt + 1), n - j_opt, cc, dd, rr, ss, w, state, 0, te);
        }
    }
}

/*
  If value for 'direction' is '1', implementation works for forward of sequences.
  If value for 'direction' is '-1', implementation works for reverse of sequences.
*/
double cost_only_in_linear_space(char *a, char *b, int m, int n, double *ccc, double *ddd, double **w, double g, int direction)
{
    double e, c, s, t;
    int i, j, ai, bj;

    //Intialize
    ddd[0] = ccc[0] = 0;
    t = G;
    for(j = 1; j <= n; j++)
    {
        ccc[j] = t = t + H;
        ddd[j] = t + G;
    }
    t = g;

    //Update cost with dynamic programming (in linear space)
    for(i = 1; i <= m; i++)
    {
        s = ccc[0];
        ddd[0] = ccc[0] = c = t = t + H;
        e = t + G;
        for(j = 1; j <= n; j++)
        {
            //getting index of element, based on forward/reverse implementation
            if(direction == 1)
            {
                ai = a[i-1];
                bj = b[j-1];
            }
            else
            {
                ai = a[m-i];
                bj = b[n-j];
            }

            e = min(e, c + G) + H;
            ddd[j] = min(ddd[j], ccc[j] + G) + H;
            c = min(min(ddd[j], e), s + w[ai][bj]);
            s = ccc[j];
            ccc[j] = c;
        }
    }

    //Test purpose return values
    if(direction == 1)
        return ccc[n];
    else
        return ccc[0];
}

//Encoding the operations in the 'state' array
void align_place(char *state, int value, char a, char b)
{
    int i, k;
    if(value == 0)
    {
        //Store '!', when both symbols are same
        if(a == b)
            state[align_length++] = '!';
        //Store '|', when both symbol is replaced
        else
            state[align_length++] = '|';
    }
    //Store 'x', when delete operation occurs
    else if(value < 0)
    {
        k = -1 * value;
        for(i = 0; i < k; i++)
            state[align_length++] = 'x';
    }
    //Store '-', when insert operation occurs
    else
        for(i = 0; i < value; i++)
            state[align_length++] = '-';
}

//Print the aligmnemt using the encoded data in the array 'state'
//Also manage the display of alignment, where each row wil show at most 'ALIGN_WIDTH' characaters
void print_alignment(char *a, char *b, int m, int n, char *state)
{
    int i = 0, j1 = 0, j2 = 0, k, pos;
    for(k = 0; k < align_length; k += ALIGN_WIDTH)
    {
    	pos = j1+1;
    	printf("%-12s %-7d", "Sequence1", pos);
        for(i = k; i < align_length && i < k + ALIGN_WIDTH; i++)
            //If operation is insert space is printed for 1st sequence
            if(state[i] == '-')
                printf(" ");
            else
                printf("%c", a[j1++]);
        if(pos < j1) pos = j1;
        printf("%7d\n%20s", pos, "");

        //For displaying both insert and delete operation, '-' will be printed
        for(i = k; i < align_length && i < k + ALIGN_WIDTH; i++)
            if(state[i] == 'x')
                printf("-");
            else
                printf("%c", state[i]);
        printf("\n");
        pos = j2 + 1;
        printf("%-12s %-7d", "Sequence2", pos);
        for(i = k; i < align_length && i < k + ALIGN_WIDTH; i++)
            //If operation is delete space is printed for 2nd sequence
            if(state[i] == 'x')
                printf(" ");
            else
                printf("%c", b[j2++]);
        if(pos < j2) pos = j2;
        printf("%7d\n\n",pos);
    }
}

//Calculate the function gap(k)
double gap(int k)
{
    if(k < 1)
        return 0;
    double gapCost = (G + H*k);
    //if(k % 3 == 0)
    return gapCost;
}

/*
Intialize 'w' according to need, based on given size.
By default, it has been implemented as 128*128 Matrix.
*/
void initialize_W(double **w, int size)
{
    int i, j;

    for(i = 0; i < size; i++)
    {
        for(j = 0; j < size; j++)
        {
            if(i==j)
                w[i][j] = 0.0;
            else
                w[i][j] = replaceCostFunc((char)i, (char)j);
        }
    }
}

double replaceCostFunc(char c1, char c2)
{
	if(!takenInputDummy)
	{
		printf("Give cost and reduced cost\n");
		scanf("%lf %lf", &replaceCost, &reducedCost);
		takenInputDummy = 1;
	}
	if(isPurine(c1) && isPurine(c2))
		return reducedCost;
	if(isPyrimidine(c1) && isPyrimidine(c2))
		return reducedCost;
	return replaceCost;
}

//This function is used, when it is expected that input sequence should be DNA/RNA
int is_valid_nucleotide(char *c)
{
    *c = toupper(*c);
    if(*c == 'N')
    	random_nucleotide(c);
    if(*c == 'A' || *c == 'U' || *c == 'C' || *c == 'G')
        return 1;
    return 0;
}

//If there is a 'N' in the sequence generate a random nucleotide
void random_nucleotide(char *c)
{
	printf("ERROR - Nucleotide '%c', converted to", *c);
	getchar();
	int mod = rand() % 4;
	if(mod == 0) *c = 'A';
	else if(mod == 1) *c = 'C';
	else if(mod == 2) *c = 'G';
	else if(mod == 3) *c = 'U';
	printf("'%c'\n", *c);
}

/*This function is used as a self testing module,
by calculating the cost of operations for the proposed alignment*/
double cost_calculation_from_alignment(char *state)
{
    int i;
    //char *state = "!-!!!!||||!!!!!!!!!!|--!!!!!|!!|!!||||!!-!!!!!!!!!!|!||!!!!-!!-!!!!!!!!-!!|!-!!";
    enum
    {
        insert,
        delet,
        replace
    } previous_op;

    previous_op = replace;
    double cost = 0.0;
    for(i = 0; i < align_length; i++)
    {
        if(state[i] == '|')
        {
            cost += REPLACE_COST_DUMMY;
            previous_op = replace;
        }
        else if(state[i] == 'x')
        {
            if(previous_op == delet)
                cost += H;
            else
                cost += G + H;
            previous_op = delet;
        }
        else if(state[i] == '-')
        {
            if(previous_op == insert)
                cost += H;
            else
                cost += G + H;
            previous_op = insert;
        }
        else
            previous_op = replace;
    }
    return cost;
}
